# Grégory Coutable 2022
# Partage sous licence Creative Commons :
# https://creativecommons.org/licenses/by-sa/3.0/fr/
#
# Version 1

import pygame
from pygame.locals import *
from random import randint, shuffle
#pour le rendre dispo de n'importe où
import os
pathname = os.path.dirname(__file__)
#positionne la fenètre en haut à gauche
x = 100
y = 100
os.environ['SDL_VIDEO_WINDOW_POS'] = "%d,%d" % (x,y)

class GUIdomino (object):
    
    def __init__(self):
        """
        Initialise une fenètre du jeu puissance 4 et l'affiche totalement vide.
        Cette classe ne contient aucun attribut publique.
        """
        
        #initialisation de pygame
        pygame.init()
        
        self.wPiece = 30
        self.wTapis = 1000
        self.hTapis = 800
        
        #on définit la fenêtre de base de notre jeu
        self.fenetre = pygame.display.set_mode((self.wTapis, self.hTapis))
        #un titre sur cette fenêtre
        pygame.display.set_caption("Dominos")
        #Chargement et collage du fond
        self.fenetre.fill(0x90EE90)
        
        self.sheetDominosH = pygame.image.load(os.path.join(pathname, "dominos_H.png")).convert_alpha()
        self.sheetDominosV = pygame.image.load(os.path.join(pathname, "dominos_V.png")).convert_alpha()
        self.verso = pygame.image.load(os.path.join(pathname, "verso.png")).convert_alpha()
        
        self.rectV = pygame.Rect(0, 0, self.wPiece, 2*self.wPiece)
        self.rectH = pygame.Rect(0, 0, 2*self.wPiece, self.wPiece)
        
        #positions des pièces des joueurs
        self.y0 = 50
        self.y1 = self.hTapis - self.y0 -  self.wPiece * 2
        
        #position de la pioche
        self.posPioche = 60
        self.zonePioche = pygame.Rect(0, 0, 160, 160)
        self.recTxtPioche = pygame.Rect(0, 0, 160, 30)
        #chaque pièce a une position aléatoire pour donner une impression de tas
        self.posAleaPioche = [(randint(self.posPioche - 30, self.posPioche + 30), randint(self.posPioche - 30, self.posPioche + 30)) for i in range(28)]
        self.angleAleaPioche = [randint(0, 360) for i in range(28)]
      
        #rectangle de détection du click souris
        self.extremite = [0, 0]
        self.piecesJoueur1 = []
        self.zoneJoueur2 = 0
        self.plateauVide = True
        
        #Rafraîchissement de l'écran
        pygame.display.update()
        

    def _getPieceH(self, n, m):
        """
        Renvoi l'image horizontale du domino (n, m)
        """
        return self.sheetDominosH.subsurface(pygame.Rect(2*n*self.wPiece, m*self.wPiece, 2*self.wPiece, self.wPiece))
    
    def _getPieceV(self, n, m):
        """
        Renvoi l'image verticale du domino (n, m)
        """
        return self.sheetDominosV.subsurface(pygame.Rect(n*self.wPiece, 2*m*self.wPiece, self.wPiece, 2*self.wPiece))

    def _putPremPiece(self, n):
        """
        Pose le premier double (n, n) en plein milieu de la fenètre.
        """
        rectScreen = self.fenetre.get_rect()
        self.rectV.center = rectScreen.center
        self.rectH.center = rectScreen.center
        self.dir = 'E'
        self.dernier = self.rectV
        #C'est un double !!
        self.fenetre.blit(self._getPieceV(n, n), self.rectV)

    def _putSens2(self):
        rectScreen = self.fenetre.get_rect()
        self.rectV.center = rectScreen.center
        self.rectH.center = rectScreen.center
        self.dernier = self.rectV
        self.dir = 'W'
        

    def _putPiece(self, n, m):
        if self.dir == 'E':
            if n != m:
                self.rectH.left = self.dernier.right
                self.fenetre.blit(self._getPieceH(n, m), self.rectH)
                self.dernier = self.rectH
            else:
                self.rectV.left = self.dernier.right
                self.fenetre.blit(self._getPieceV(n, m), self.rectV)
                self.dernier = self.rectV
                
        elif self.dir == 'S':
            if n != m:
                self.rectV.top = self.dernier.bottom
                self.fenetre.blit(self._getPieceV(n, m), self.rectV)
                self.dernier = self.rectV
            else:
                self.rectH.top = self.dernier.bottom
                self.fenetre.blit(self._getPieceH(n, m), self.rectH)
                self.dernier = self.rectH           

        elif self.dir == 'W':
            if n != m:
                self.rectH.right = self.dernier.left
                self.fenetre.blit(self._getPieceH(n, m), self.rectH)
                self.dernier = self.rectH
            else:
                self.rectV.right = self.dernier.left
                self.fenetre.blit(self._getPieceV(n, m), self.rectV)
                self.dernier = self.rectV

        elif self.dir == 'N':
            if n != m:
                self.rectV.bottom = self.dernier.top
                self.fenetre.blit(self._getPieceV(n, m), self.rectV)
                self.dernier = self.rectV
            else:
                self.rectH.bottom = self.dernier.top
                self.fenetre.blit(self._getPieceH(n, m), self.rectH)
                self.dernier = self.rectH

    def _tourneDroite(self):
        if self.dir == 'E':
            self.rectV.right = self.dernier.right
            self.rectH.right = self.dernier.right + self.wPiece // 2
            self.dir = 'S'

        elif self.dir == 'S':
            self.rectH.bottom = self.dernier.bottom
            self.rectV.bottom = self.dernier.bottom + self.wPiece // 2
            self.dir = 'W'

        elif self.dir == 'W':
            self.rectV.left = self.dernier.left
            self.rectH.left = self.dernier.left - self.wPiece // 2
            self.dir = 'N'

        elif self.dir == 'N':
            self.rectH.top = self.dernier.top
            self.rectV.top = self.dernier.top - self.wPiece // 2
            self.dir = 'E'
            
    def _putPioche(self, n):
        for i in range(n):
            piecePioche = pygame.transform.rotate(self.verso, self.angleAleaPioche[i%28])
            self.fenetre.blit(piecePioche, self.posAleaPioche[i%28])

    def refresh(self, g, j1, j2, n, t = "", cacheJ2 = False):
        """
        Cette méthode rafraichie l'affichage du jeu de domino conformément à la grille passée en argument.
        
        j1 et j2 sont des objets indexables, dont les éléments sont de la forme (n, m).
            - n et m sont deux entiers de 0 à 6 représentant le nombre de points de chaque coté du domino.
        D'autre  part j1 et j2 doivent disposer d'une méthode __len__ pour répondre a la fonction len().
        Il peut donc s'agir de listes, de tuples ou de tout autre objets répondants a ces critères.
        
        cacheJ2 est un booléen. Si cet argument est vrai, toutes les dominos de l'ordinateur seront
        affichés retournés.
        
        n est le nombe de dominos dans la pioche.
        
        t est un texte à afficher à destination du joueur.
        """
        
        #réinitialisation du fond
        self.fenetre.fill(0x90EE90)
        
        
        
        if g:
            self.plateauVide = False
            
            pivot = self._cherchePivot(g)
            self._putPremPiece(g[pivot][0])
            
            #séparation de la liste en deux autour du pivot : vers l'est, vers l'ouest
            E = g[pivot + 1:]
            W = g[:pivot]
            W.reverse()
            lstEW = [E, W]
            
            for j in range(2):
                nbVirage = 0
                for i in range(len(lstEW[j])):
                    if self._testVirage() and lstEW[j][i][0] != lstEW[j][i][1]:
                        nbVirage += 1
                        self._tourneDroite()
                    if nbVirage == 2:
                        self._putPiece(lstEW[j][i][1], lstEW[j][i][0])
                    else:    
                        self._putPiece(lstEW[j][i][0], lstEW[j][i][1])
                #on mémorise les deux extrémités pour détection du click
                self.extremite[j] = self.dernier.copy()
                    
                self._putSens2()
        
        else:
            self.plateauVide = True

        delta = 10
        rectJ1 = pygame.Rect((self.wTapis - (self.wPiece * len(j1) + delta * (len(j1) - 1)))//2, self.y1, self.wPiece, 2*self.wPiece)
        
        self.piecesJoueur1 = []
        for u in j1:
            self.fenetre.blit(self._getPieceV(u[0], u[1]), rectJ1)
            #On mémorise les emplacement de chaque pièce posée
            self.piecesJoueur1.append(rectJ1.copy())
            rectJ1.left += self.wPiece + delta
            
        rectJ2 = pygame.Rect((self.wTapis - (self.wPiece * len(j2) + delta * (len(j2) - 1)))//2, self.y0, self.wPiece, 2*self.wPiece)
        
        self.zoneJoueur2 = pygame.Rect((self.wTapis - (self.wPiece * len(j2) + delta * (len(j2) - 1)))//2, self.y0, (self.wPiece * len(j2) + delta * (len(j2) - 1)), 2*self.wPiece)
        
        for u in j2:
            if cacheJ2:
                self.fenetre.blit(self.verso, rectJ2)
            else:
                self.fenetre.blit(self._getPieceV(u[0], u[1]), rectJ2)
            rectJ2.left += self.wPiece + delta           

        #zone de texte
        #Une police
        police = pygame.font.SysFont("Arial Black",28)
        #un texte
        texte = police.render(t, True, pygame.Color("#FFFFFF"))
        #Une surface pour centrer le texte sur le bandeau du haut
        surfText = pygame.Rect(0, self.hTapis - 250, self.wTapis, 100)
        #pour centrer le texte
        rectTexte = texte.get_rect()
        rectTexte.center = surfText.center
        self.fenetre.blit(texte, rectTexte)
        
        texte = police.render("PIOCHE", True, pygame.Color("#FFFF00"))
        rectTexte = texte.get_rect()
        rectTexte.center = self.recTxtPioche.center
        self.fenetre.blit(texte, rectTexte)
        
        self._putPioche(n)
        
        #Rafraîchissement de l'écran
        pygame.display.update()


    def _testVirage(self):
        """
        Cette méthode vérifie si il est necessaire de tourner pour ne pas que la ligne de domino ne sorte de l'écran.
        Renvoi un booléen.
        """
        # marge au bord de l'écran doite/gauche et haut/bas
        margeDG = 100
        margeHB = 300
        if self.dir == 'W':
            if self.dernier.left < margeDG:
                return True
            return False

        elif self.dir == 'S':
            if self.dernier.bottom > self.hTapis - margeHB:
                return True
            return False
        
        elif self.dir == 'E':
            if self.dernier.right > self.wTapis - margeDG:
                return True
            return False

        elif self.dir == 'N':
            if self.dernier.top < margeHB:
                return True
            return False
        

    @staticmethod
    def _cherchePivot(g):
        """
        Cherche le double le plus près du milieu de la liste de pièce g. Si pas de double : levée d'une exception.
        """
        m = len(g)//2
        
        for k in range(len(g)):
            if m + k < len(g) and g[m + k][0] == g[m + k][1]:
                return m + k
            if m - k >= 0 and g[m - k][0] == g[m - k][1]:
                return m - k
        raise "La liste de pièce doit contenir obligatoirement au moins un double"
                
            
            
            
    def messCentre(self, mess):
        """
        Cette méthode permet d'afficher mess en plein écran au milieu.
        """
        police = pygame.font.Font(os.path.join(pathname, "led.ttf"),100)
        texte = police.render(mess, True, pygame.Color("#FFFF00"))
        #pour centrer le texte
        rectTexte = texte.get_rect()
        rectTexte.center = self.fenetre.get_rect().center
        rectTexte.bottom -= 82
        self.fenetre.blit(texte,rectTexte)
        #Rafraîchissement de l'écran
        pygame.display.update()

    def waitClick(self):
        """
        Cette méthode attend l'action d'un joueur. Elle gère trois type d'actions :
            - demande fermeture de la fenètre : fermeture propre de la fenètre pygame et fin du programme python.
            - click sur la fenetre : retourne  0 si click sur les cartes du croupier et 1 si click sur les cartes du joueur.
            - appui sur des touches (uniquement y, n, s, r, c, d, fleche droite : retourne la lettre correspondante.
        Une fois exécutée, on ne peut sortir de cette méthode que par l'une de ces trois actions.
        """
        while True:
            #Limitation de vitesse de la boucle
            pygame.time.Clock().tick(30)
            
            for event in pygame.event.get():    #Attente des événements
                if event.type == QUIT:
                    pygame.quit()
                    exit()
                    
                if event.type == MOUSEBUTTONDOWN:
                    if event.button == 1:   #Si clic gauche
                        #détection des extrémités
                        if not self.plateauVide and self.extremite[0].collidepoint(event.pos):
                            return '#D'
                        if not self.plateauVide and self.extremite[1].collidepoint(event.pos):
                            return '#G'
                        if self.zonePioche.collidepoint(event.pos):
                            return '#P'
                        if self.zoneJoueur2.collidepoint(event.pos):
                            return '#A'              
                        for i in range(len(self.piecesJoueur1)):
                            if self.piecesJoueur1[i].collidepoint(event.pos):
                                return i                            
                    
                if event.type == KEYDOWN:
                    touches = {K_RIGHT : '_R', K_LEFT : '_L', K_UP : '_U', K_DOWN : '_D', K_RETURN : '_E', K_BACKSPACE : '_B', K_ESCAPE : '_S'}
                    touche = event.key
                    if touche in touches:
                        return touches[touche]
                    return event.unicode  
                    
                         
if __name__ == "__main__":
    import time
    #instanciation d'une fenètre de domino
    GUI = GUIdomino()
    
    g = [[4, 5], [5, 5], [5, 3], [3, 2], [2, 2], [2, 5], [5, 6], [6, 6], [6, 3], [3, 5], [5, 5], [5, 0], [0, 6], [6, 4], [4, 4], [4, 3], [3, 3], [3, 1], [1, 1], [1, 0], [0, 4], [4, 2], [2, 6], [6, 1], [1, 2], [2, 0], [0, 0], [0, 3], [3, 1], [1, 2]]
    
  
    j1 = [(randint(0,6), randint(0,6)) for i in range(10)]
    j2 = [(randint(0,6), randint(0,6)) for i in range(5)]

    GUI.refresh(g, j1, j2, 5, "Score Joeur 1 : 256   Score joueur 2 : 41", True)
    time.sleep(2)

    GUI.messCentre("GAME OVER")
    time.sleep(2)
   
    for i in range(29):
       GUI.refresh(g, j1, j2, i, "Modification de la pioche", False)
       time.sleep(0.1)

    GUI.refresh(g, j1, j2, 28, "A vous de jouer !!", True)
    

    
    while True :
        print(GUI.waitClick())

